; ==================================================================================================
; Title:   XML3DContentHandler.inc
; Author:  G. Friedrich
; Version: 1.0.0
; Purpose: Support of COM XML3DContentHandler interface for ObjAsm32.
; Links:   http://msdn2.microsoft.com/en-us/library/ms767671.aspx
;          http://www.perfectxml.com/Articles/MSXML/September02.asp?pg=1
; Notes:   Version 1.0.0, January 2007
;            - First release.
; ==================================================================================================


Object XML3DContentHandler, XML3DContentHandlerID, Primer
  RedefineMethod    Done
  RedefineMethod    Init,                   Pointer, Pointer, Pointer, Pointer, Pointer, Pointer
  StaticMethod      GetCurrentPath

  ;Don't change the order of the interface methods!
  InterfaceMethod   QueryInterface,         Pointer, Pointer
  InterfaceMethod   AddRef
  InterfaceMethod   Release
  InterfaceMethod   putDocumentLocator,     Pointer
  InterfaceMethod   startDocument
  InterfaceMethod   endDocument
  InterfaceMethod   startPrefixMapping,     Pointer, dword, Pointer, dword
  InterfaceMethod   endPrefixMapping,       Pointer, dword
  InterfaceMethod   startElement,           Pointer, dword, Pointer, dword, Pointer, dword, Pointer
  InterfaceMethod   endElement,             Pointer, dword, Pointer, dword, Pointer, dword
  InterfaceMethod   characters,             Pointer, dword
  InterfaceMethod   ignorableWhitespace,    Pointer, dword
  InterfaceMethod   processingInstruction,  Pointer, dword, Pointer, dword
  InterfaceMethod   skippedEntity,          Pointer, dword

  ;The owner manages the following objects
  DefineVariable  pPositionStorageColl,     Pointer,      NULL    
  DefineVariable  pNormalStorageColl,       Pointer,      NULL
  DefineVariable  pIndexStorageColl,        Pointer,      NULL
  DefineVariable  pMaterialColl,            Pointer,      NULL

  DefineVariable  pDirect3D,                Pointer,      NULL
  DefineVariable  dCurrPathCharCount,       dword,        0
  DefineVariable  pMaterial,                Pointer,      NULL
  DefineVariable  pPositionStorage,         Pointer,      NULL
  DefineVariable  pNormalStorage,           Pointer,      NULL
  DefineVariable  pIndexStorage,            Pointer,      NULL

  Embed StrColl, DataCollection

ObjectEnd



; ==================================================================================================

if IMPLEMENT

; ==================================================================================================
;    XML3DContentHandler implementation
; ==================================================================================================

DebugCountedUnicodeString macro UStrName:req
    mov eax, @CatStr(<d>, UStrName, <Len>)
    .if eax > (sizeof bBuffer/2 - 1)
      mov eax, (sizeof bBuffer/2 - 1)
    .endif
    invoke UStrLCopy, addr bBuffer, @CatStr(<p>, UStrName), eax
    DbgUStr addr bBuffer
endm

; 
; Method:    XML3DContentHandler.Done
; Purpose:   Finalizes the object.
; Arguments: None.
; Return:    Nothing.

Method XML3DContentHandler.Done, uses esi
    SetObject esi
    OCall [esi].StrColl::DataCollection.Done
    ACall esi.Done
MethodEnd

; 
; Method:    XML3DContentHandler.Init
; Purpose:   Inmitializes the object.
; Arguments: Arg1: -> Owner object.
;            Arg2: -> Direct3D object.
;            Arg3: -> owners PositionStorage Collection.
;            Arg4: -> owners NormalStorage Collection.
;            Arg5: -> owners IndexStorage Collection.
;            Arg6: -> owners Material Collection.
; Return:    Nothing.

Method XML3DContentHandler.Init, uses esi, pOwner:Pointer, pDirect3D:Pointer, \
                        pPositionStorageColl:Pointer, pNormalStorageColl:Pointer, \
                        pIndexStorageColl:Pointer, pMaterialColl:Pointer
    SetObject esi
    ACall esi.Init, pOwner
    mov [esi].dCurrPathCharCount, 0
    m2m [esi].pDirect3D, pDirect3D
    m2m [esi].pPositionStorageColl, pPositionStorageColl
    m2m [esi].pNormalStorageColl, pNormalStorageColl
    m2m [esi].pIndexStorageColl, pIndexStorageColl
    m2m [esi].pMaterialColl, pMaterialColl
    OCall [esi].StrColl::DataCollection.Init, esi, 10, 10, COLLECTIONMAXSIZE
MethodEnd

; 
; Method:    XML3DContentHandler.GetCurrentPath
; Purpose:   Returns the current path into the XML file.
; Arguments: None.
; Return:    eax -> string. Must be disposed by the caller.

AddString proc pUStr:Pointer, pCurrentPointer:Pointer
    mov eax, pCurrentPointer
    mov ecx, [eax]
    invoke UStrECopy,ecx, pUStr
    mov word ptr [eax], "|"                 ;Add separator
    add eax, 2
    mov ecx, pCurrentPointer
    mov [ecx], eax
    ret
AddString endp

Method XML3DContentHandler.GetCurrentPath, uses esi
    local dCurrentPosition:dword

    SetObject esi
    invoke UStrAlloc, [esi].dCurrPathCharCount
    .if eax != NULL
      push eax
      mov dCurrentPosition, eax
      OCall [esi].StrColl::DataCollection.ForEach, addr AddString, addr dCurrentPosition
      mov eax, [esi].dCurrPathCharCount
      shl eax, 1
      add eax, [esp]                        ;Remove last separator
      sub eax, 2
      mov word ptr [eax], 0
      pop eax
    .else
      OCall esi.ErrorReport, NULL, XML_OUT_OF_MEMORY
      xor eax, eax
    .endif
MethodEnd

; 
; Method:    XML3DContentHandler.QueryInterface
; Purpose:   Checks if the requested interface is supported. If it is, a pointer to a pointer
;            to the interface is returned and the reference count is incremented
; Arguments: Arg1: Pointer to the requested interface GUID.
;            Arg2: Pointer to the variable that will hold the interface pointer.
; Return:    HRESULT.

Method XML3DContentHandler.QueryInterface,, pReqIID:Pointer, ppInterface:Pointer
    mov eax, S_OK
MethodEnd, eax

; 
; Method:    XML3DContentHandler.AddRef
; Purpose:   Increments the internal interface reference count.
; Arguments: None.
; Return:    Current reference count.

Method XML3DContentHandler.AddRef
    mov eax, S_OK
MethodEnd, eax

; 
; Method:    XML3DContentHandler.Release
; Purpose:   Decrements the internal interface reference count.
; Arguments: None.
; Return:    (optional) Current reference count.

Method XML3DContentHandler.Release
    mov eax, S_OK
MethodEnd, eax

; 
; Method:    XML3DContentHandler.putDocumentLocator
; Purpose:   Sets a Locator interface.
; Arguments: Arg1: -> Locator interface
; Return:    HRESULT.

Method XML3DContentHandler.putDocumentLocator,, pLocator:Pointer
    mov eax, S_OK
MethodEnd, eax

; 
; Method:    XML3DContentHandler.startDocument
; Purpose:   Receives notification of the beginning of a document.
;            The reader invokes the startDocument method only once.
; Arguments: None.
; Return:    HRESULT.

Method XML3DContentHandler.startDocument
;    DbgText "XML3DContentHandler.startDocument"
    mov eax, S_OK
MethodEnd, eax

; 
; Method:    XML3DContentHandler.endDocument
; Purpose:   Receives notification of the end of a document.
;            The reader invokes the endDocument method only once. The endDocument method is the
;            last method invoked during the parse. The reader does not invoke this method until it
;            has either abandoned parsing upon encountering an unrecoverable error or has reached
;            the end of the document.
; Arguments: None.
; Return:    HRESULT.

Method XML3DContentHandler.endDocument
    mov eax, S_OK
MethodEnd, eax

; 
; Method:    XML3DContentHandler.startPrefixMapping
; Purpose:   Begins the scope of the prefix-URI namespace mapping.
; Arguments: Arg1: -> UNICODE prefix being mapped.
;            Arg2: Prefix character count.
;            Arg3: -> UNICODE namespace URI to which the prefix is mapped.
;            Arg4: Namespace character count.
; Return:    HRESULT.

Method XML3DContentHandler.startPrefixMapping,, pPrefix:Pointer, dPrefixLen:dword, \
                                                pNamespace:Pointer, dNamespaceLen:dword
    mov eax, S_OK
MethodEnd, eax

; 
; Method:    XML3DContentHandler.endPrefixMapping
; Purpose:   Ends the scope of the prefix-URI namespace mapping.
; Arguments: Arg1: -> UNICODE prefix being mapped.
;            Arg2: Prefix character count.
; Return:    HRESULT.

Method XML3DContentHandler.endPrefixMapping,, pPrefix:Pointer, dPrefixLen:dword
    mov eax, S_OK
MethodEnd, eax

; 
; Method:    XML3DContentHandler.startElement
; Purpose:   The reader invokes this method at the start of every element in the XML document.
; Arguments: Arg1: -> Namespace URI.
;            Arg2: Namespace URI character count.
;            Arg3: -> UNICODE local name.
;            Arg4: Local name character count.
;            Arg5: -> UNICODE QName.
;            Arg6: QName character count.
;            Arg7: -> Attributes attached to the element.
; Return:    HRESULT.

ReadNumber macro pNumber:req, dCount:req
    mov edi, pNumber
    mov eax, dCount
    .if eax > (sizeof bNumText/2 - 1)
      mov eax, (sizeof bNumText/2 - 1)
    .endif
    lea ecx, bNumText
    push ecx
    invoke UStrLCopy, ecx, pNumber, eax
    invoke UStr2Str, [esp + 4], [esp]
    call StrToSt0
endm

ReadNumberArray macro dAttrIndex:req, Storage:req, IsInteger:req
    lea ecx, Storage
    mov [ecx].DwordCollection.dDuplicates, TRUE
    ICall pAttributes::ISAXAttributes.getValue, dAttrIndex, addr pAttrText, addr dAttrTextLen

    ;Trim ending spaces from attribute string
    mov edi, dAttrTextLen
    dec edi
    shl edi, 1
    add edi, pAttrText                                ;edi -> to last UNICODE character
    xor ebx, ebx
    .while ebx < dAttrTextLen
      .break .if word ptr [edi] != " "
      dec dAttrTextLen
      sub edi, 2
      inc ebx
    .endw

    ;Read each subnumber and add it to the collection
    xor ebx, ebx
    mov edi, pAttrText
    .while ebx <= dAttrTextLen
      .if word ptr [edi] == " " || word ptr [edi] == "," || ebx == dAttrTextLen
        mov eax, edi
        sub eax, pAttrText
        shr eax, 1
        .if eax > (sizeof bNumText/2 - 1)
          mov eax, (sizeof bNumText/2 - 1)
        .endif
        lea ecx, bNumText
        push ecx
        invoke UStrLCopy, ecx, pAttrText, eax
        invoke UStr2Str, [esp + 4], [esp]
        call StrToSt0
        if IsInteger eq FALSE
          fstp r4Value
        else
          fistp r4Value
        endif
        OCall Storage::DwordCollection.Insert, r4Value
        inc edi
        inc ebx
        inc edi
        mov pAttrText, edi
      .else
        inc edi
        inc ebx
        inc edi
      .endif
    .endw
endm

Method XML3DContentHandler.startElement, uses ebx edi esi, pNamespaceURI:Pointer, dNamespaceURILen:dword, \
         pLocalName:Pointer, dLocalNameLen:dword, pQName:Pointer, dQNameLen:dword, pAttributes:Pointer
    local dIndex:dword
    local bNumText[100]:byte, r4Value:real4
    local pAttrText:Pointer, dAttrTextLen:dword


;    DbgText "XML3DContentHandler.startElement"
    SetObject esi
    invoke UStrLNew, pLocalName, dLocalNameLen
    OCall [esi].StrColl::DataCollection.Insert, eax
    mov eax, dLocalNameLen
    inc eax
    add [esi].dCurrPathCharCount, eax

    OCall esi.GetCurrentPath
    .if eax != NULL
      push eax

      ;Find the Surface node
      invoke UStrIComp, [esp + 4], $OfsCUStr("Model_3dxml|GeometricRepresentationSet|Representation|AssociatedXML|Rep")
      .if eax == 0
        MemAlloc sizeof D3DMATERIAL9, MEM_INIT_ZERO
        or eax, eax
        jz @@OUTOFMEM
        mov [esi].pMaterial, eax
        ;Fill with std color
        fmov [eax].D3DMATERIAL9.Diffuse.r, 1.0
        fmov [eax].D3DMATERIAL9.Diffuse.g, 1.0
        fmov [eax].D3DMATERIAL9.Diffuse.b, 1.0
        fmov [eax].D3DMATERIAL9.Diffuse.a, 1.0
        
        New DwordCollection
        or eax, eax
        jz @@OUTOFMEM
        mov [esi].pPositionStorage, eax
        mov [eax].DwordCollection.dDuplicates, TRUE
        OCall eax::DwordCollection.Init, esi, 300, 300, COLLECTIONMAXSIZE

        New DwordCollection
        or eax, eax
        jz @@OUTOFMEM
        mov [esi].pNormalStorage, eax
        mov [eax].DwordCollection.dDuplicates, TRUE
        OCall eax::DwordCollection.Init, esi, 300, 300, COLLECTIONMAXSIZE

        New DwordCollection
        or eax, eax
        jz @@OUTOFMEM
        mov [esi].pIndexStorage, eax
        mov [eax].DwordCollection.dDuplicates, TRUE
        OCall eax::DwordCollection.Init, esi, 300, 300, COLLECTIONMAXSIZE
        
        jmp @@Exit
@@OUTOFMEM:
        OCall esi.ErrorReport, NULL, D3D_OUT_OF_MEMORY
        jmp @@Exit
      .endif

      ;Find the VertexBuffer node
      invoke UStrIComp, [esp + 4], $OfsCUStr("Model_3dxml|GeometricRepresentationSet|Representation|AssociatedXML|Rep|VertexBuffer")
      .if eax == 0
        ;Vertex information found
        .if pAttributes != NULL
          ICall pAttributes::ISAXAttributes.getIndexFromName, NULL, 0, $OfsCUStr("positions"), 9, addr dIndex
          .if SUCCEEDED(eax)
            ;Positions vectors found - we will try to read it
            ReadNumberArray dIndex, [esi].pPositionStorage, FALSE
            ICall pAttributes::ISAXAttributes.getIndexFromName, NULL, 0, $OfsCUStr("normals"), 7, addr dIndex
            .if SUCCEEDED(eax)
              ;Normal vectors found - we will try to read it
              ReadNumberArray dIndex, [esi].pNormalStorage, FALSE
            .endif
          .endif
        .endif
        jmp @@Exit
      .endif

      invoke UStrIComp, [esp + 4], $OfsCUStr("Model_3dxml|GeometricRepresentationSet|Representation|AssociatedXML|Rep|Faces|Face")
      .if eax == 0
        ;Face information found
        .if pAttributes != NULL
          ICall pAttributes::ISAXAttributes.getIndexFromName, NULL, 0, $OfsCUStr("triangles"), 9, addr dIndex
          .if SUCCEEDED(eax)
            ;Face indices found - we will try to read it
            ReadNumberArray dIndex, [esi].pIndexStorage, TRUE
          .endif
        .endif
        jmp @@Exit
      .endif

      invoke UStrIComp, [esp + 4], $OfsCUStr("Model_3dxml|GeometricRepresentationSet|Representation|AssociatedXML|Rep|Faces|Face|SurfaceAttributes|Color")
      .if eax == 0
        ;Color information found
        .if pAttributes != NULL
          ICall pAttributes::ISAXAttributes.getIndexFromName, NULL, 0, $OfsCUStr("red"), 3, addr dIndex
          .if SUCCEEDED(eax)
            ICall pAttributes::ISAXAttributes.getValue, dIndex, addr pAttrText, addr dAttrTextLen
            ReadNumber pAttrText, dAttrTextLen
            mov eax, [esi].pMaterial
            fst [eax].D3DMATERIAL9.Ambient.r
            fstp [eax].D3DMATERIAL9.Diffuse.r
          .endif
          ICall pAttributes::ISAXAttributes.getIndexFromName, NULL, 0, $OfsCUStr("green"), 5, addr dIndex
          .if SUCCEEDED(eax)
            ICall pAttributes::ISAXAttributes.getValue, dIndex, addr pAttrText, addr dAttrTextLen
            ReadNumber pAttrText, dAttrTextLen
            mov eax, [esi].pMaterial
            fst [eax].D3DMATERIAL9.Ambient.g
            fstp [eax].D3DMATERIAL9.Diffuse.g
          .endif
          ICall pAttributes::ISAXAttributes.getIndexFromName, NULL, 0, $OfsCUStr("blue"), 4, addr dIndex
          .if SUCCEEDED(eax)
            ICall pAttributes::ISAXAttributes.getValue, dIndex, addr pAttrText, addr dAttrTextLen
            ReadNumber pAttrText, dAttrTextLen
            mov eax, [esi].pMaterial
            fst [eax].D3DMATERIAL9.Ambient.b
            fstp [eax].D3DMATERIAL9.Diffuse.b
          .endif
          ICall pAttributes::ISAXAttributes.getIndexFromName, NULL, 0, $OfsCUStr("alpha"), 5, addr dIndex
          .if SUCCEEDED(eax)
            ICall pAttributes::ISAXAttributes.getValue, dIndex, addr pAttrText, addr dAttrTextLen
            ReadNumber pAttrText, dAttrTextLen
            mov eax, [esi].pMaterial
            fst [eax].D3DMATERIAL9.Ambient.a
            fstp [eax].D3DMATERIAL9.Diffuse.a
          .endif

        .endif
        jmp @@Exit
      .endif

@@Exit:
      call UStrDispose

    .endif
    mov eax, S_OK
MethodEnd, eax

; 
; Method:    XML3DContentHandler.endElement
; Purpose:   The reader invokes this method at the end of every element in the XML document.
; Arguments: Arg1: -> Namespace URI.
;            Arg2: Namespace URI character count.
;            Arg3: -> UNICODE local name.
;            Arg4: Local name character count.
;            Arg5: -> UNICODE QName.
;            Arg6: QName character count.
; Return:    HRESULT.

Method XML3DContentHandler.endElement, uses ebx esi, pNamespaceURI:Pointer, dNamespaceURILen:dword, \
         pLocalName:Pointer, dLocalNameLen:dword, pQName:Pointer, dQNameLen:dword
    local dIndex:Pointer, pCurrPosition:Pointer, pCurrNormal:Pointer
    local pBuffer:Pointer, dCount:dword

;    DbgText "XML3DContentHandler.endElement"
    SetObject esi

    OCall esi.GetCurrentPath
    .if eax != NULL
      push eax

      ;Find the Surface node
      invoke UStrIComp, [esp + 4], $OfsCUStr("Model_3dxml|GeometricRepresentationSet|Representation|AssociatedXML|Rep")
      .if eax == 0
        ;Time to store the info we got
        OCall [esi].pPositionStorageColl::Collection.Insert, [esi].pPositionStorage
        OCall [esi].pNormalStorageColl::Collection.Insert, [esi].pNormalStorage
        OCall [esi].pIndexStorageColl::Collection.Insert, [esi].pIndexStorage
        OCall [esi].pMaterialColl::DataCollection.Insert, [esi].pMaterial
      .endif

      call UStrDispose
    .endif

    lea ebx, [esi].StrColl
    mov eax, [ebx].DataCollection.dCount
    .if eax > 0
      dec eax
      mov dIndex, eax
      OCall ebx::DataCollection.ItemAt, eax
      invoke UStrLComp, eax, pLocalName, dLocalNameLen
      .if eax == 0
        OCall ebx::DataCollection.DeleteAt, dIndex
        mov eax, dLocalNameLen
        inc eax
        sub [esi].dCurrPathCharCount, eax
      .else
        OCall esi.ErrorReport, NULL, XML_WRONG_TAG
      .endif
    .else
      OCall esi.ErrorReport, NULL, XML_WRONG_TAG
    .endif

    mov eax, S_OK
MethodEnd, eax

; 
; Method:    XML3DContentHandler.characters
; Purpose:   Receives notification of character data.
; Arguments: Arg1: -> UNICODE characters.
;            Arg2: Number of characters.
; Return:    HRESULT.

Method XML3DContentHandler.characters,, pChars:Pointer, dCharsLen:dword
    mov eax, S_OK
MethodEnd, eax

; 
; Method:    XML3DContentHandler.ignorableWhitespace
; Purpose:   This method is not called in the current implementation because the Simple API
;            for XML (SAX2) implementation is nonvalidating.
; Arguments: Arg1: -> UNICODE characters.
;            Arg2: Character count.
; Return:    HRESULT.

Method XML3DContentHandler.ignorableWhitespace,, pChars:Pointer, dCharsLen:dword
    mov eax, S_OK
MethodEnd, eax

; 
; Method:    XML3DContentHandler.processingInstruction
; Purpose:   Receives notification of a processing instruction.
;            The reader invokes the processingInstruction method once for each processing
;            instruction found. Processing instructions can occur before or after the main document element.
; Arguments: Arg1: -> UNICODE Target.
;            Arg2: Target character count.
;            Arg3: -> UNICODE Data.
;            Arg4: Data character count.
; Return:    HRESULT.

Method XML3DContentHandler.processingInstruction,, pTarget:Pointer, dTargetLen:dword, \
                                                   pData:Pointer, dDataLen:dword
    mov eax, S_OK
MethodEnd, eax

; 
; Method:    XML3DContentHandler.skippedEntity
; Purpose:   Receives notification of a skipped entity.
;            The skippedEntity method is invoked for each entity skipped. The reader may skip
;            entities if they have not been seen in the declarations. An example of this is when an
;            entity is declared in an external DTD subset.
; Arguments: Arg1: -> UNICODE Name.
;            Arg2: Name character count.
; Return:    HRESULT.

Method XML3DContentHandler.skippedEntity,, pName:Pointer, dNameLen:dword
    mov eax, S_OK
MethodEnd, eax


endif
